>From bad5edea46eeb83cc31217049265c69751267dfc Mon Sep 17 00:00:00 2001
From: Joseph Tulou <brezhoneg1@yahoo.fr>
Date: Fri, 12 Dec 2008 01:23:57 +0100
Subject: [PATCH] fixes for media_list and media_list_player (libvlc api)

---
 include/vlc/libvlc_events.h     |   26 ++-
 include/vlc/libvlc_media_list.h |   26 +++-
 src/control/core.c              |    3 +
 src/control/event.c             |    6 +-
 src/control/libvlc_internal.h   |    7 +-
 src/control/media_list.c        |    2 +-
 src/control/media_list_path.h   |   10 +-
 src/control/media_list_player.c |  397 +++++++++++++++++++++++++++++++--------
 src/control/media_player.c      |    2 +-
 src/libvlc.sym                  |    2 +
 10 files changed, 382 insertions(+), 99 deletions(-)

diff --git a/include/vlc/libvlc_events.h b/include/vlc/libvlc_events.h
index 5a0984a..5a788dd 100644
--- a/include/vlc/libvlc_events.h
+++ b/include/vlc/libvlc_events.h
@@ -79,7 +79,7 @@ typedef enum libvlc_event_type_t {
     libvlc_MediaListViewItemDeleted,
     libvlc_MediaListViewWillDeleteItem,
 
-    libvlc_MediaListPlayerPlayed,
+    libvlc_MediaListPlayerItemPlayed,
     libvlc_MediaListPlayerNextItemSet,
     libvlc_MediaListPlayerStopped,
 
@@ -88,6 +88,10 @@ typedef enum libvlc_event_type_t {
 
     libvlc_MediaPlayerTitleChanged,
     libvlc_MediaPlayerSnapshotTaken,
+
+    libvlc_MediaListPlayerStarted,
+    libvlc_MediaListPlayerEndReached
+
     /* New event types HERE */
 } libvlc_event_type_t;
 
@@ -151,6 +155,10 @@ typedef struct libvlc_event_t
         {
             libvlc_time_t new_pausable;
         } media_player_pausable_changed;
+        struct
+        {
+             char* psz_filename ;
+        } media_player_snapshot_taken ;
 
         /* media list */
         struct
@@ -196,6 +204,16 @@ typedef struct libvlc_event_t
             int index;
         } media_list_view_will_delete_item;
 
+        /* media list player */
+        struct
+        {
+            libvlc_media_t * item;
+        } media_item_played;
+        struct
+        {
+            libvlc_media_t * item;
+        } media_next_item_set;
+
         /* media discoverer */
         struct
         {
@@ -206,12 +224,6 @@ typedef struct libvlc_event_t
             void * unused;
         } media_media_discoverer_ended;
 
-        /* snapshot taken */
-        struct
-        {
-             char* psz_filename ;
-        } media_player_snapshot_taken ;
-
     } u;
 } libvlc_event_t;
 
diff --git a/include/vlc/libvlc_media_list.h b/include/vlc/libvlc_media_list.h
index 94a4f7b..6572c39 100644
--- a/include/vlc/libvlc_media_list.h
+++ b/include/vlc/libvlc_media_list.h
@@ -208,7 +208,7 @@ VLC_PUBLIC_API void
  * Get a flat media list view of media list items
  *
  * \param p_ml a media list instance
- * \param p_ex an excpetion instance
+ * \param p_ex an exception instance
  * \return flat media list view instance
  */
 VLC_PUBLIC_API libvlc_media_list_view_t *
@@ -392,6 +392,18 @@ VLC_PUBLIC_API void
     libvlc_media_list_player_release( libvlc_media_list_player_t * p_mlp );
 
 /**
+ * Get media player in media_list_player with this instance.
+ *
+ * \param p_mlp media list player instance
+ * \param p_e initialized exception instance
+ * \return media player instance
+ */
+VLC_PUBLIC_API libvlc_media_player_t *
+    libvlc_media_list_player_get_media_player(
+                                     libvlc_media_list_player_t * p_mlp,
+                                     libvlc_exception_t * p_e );
+
+/**
  * Replace media player in media_list_player with this instance.
  *
  * \param p_mlp media list player instance
@@ -493,6 +505,18 @@ VLC_PUBLIC_API void
 
 /* NOTE: shouldn't there also be a libvlc_media_list_player_prev() */
 
+/**
+ * Get libvlc_event_manager from this media list player instance.
+ * The p_event_manager is immutable, so you don't have to hold the lock
+ *
+ * \param p_mlp a media list player instance
+ * \param p_ex an exception instance
+ * \return libvlc_event_manager
+ */
+VLC_PUBLIC_API libvlc_event_manager_t *
+    libvlc_media_list_player_event_manager( libvlc_media_list_player_t *,
+                                            libvlc_exception_t * );
+
 /** @} media_list_player */
 
 /** @} media_list */
diff --git a/src/control/core.c b/src/control/core.c
index c31e264..c82e03c 100644
--- a/src/control/core.c
+++ b/src/control/core.c
@@ -43,6 +43,9 @@ void libvlc_exception_init( libvlc_exception_t *p_exception )
 
 void libvlc_exception_clear( libvlc_exception_t *p_exception )
 {
+    if ( !p_exception )
+        return;
+
     if( p_exception->psz_message != nomemstr )
         free( p_exception->psz_message );
     p_exception->psz_message = NULL;
diff --git a/src/control/event.c b/src/control/event.c
index d7f0b41..5dcdec3 100644
--- a/src/control/event.c
+++ b/src/control/event.c
@@ -253,6 +253,7 @@ static const char event_type_to_name[][35] =
     EVENT(libvlc_MediaPlayerPositionChanged),
     EVENT(libvlc_MediaPlayerSeekableChanged),
     EVENT(libvlc_MediaPlayerPausableChanged),
+    EVENT(libvlc_MediaPlayerSnapshotTaken),
 
     EVENT(libvlc_MediaListItemAdded),
     EVENT(libvlc_MediaListWillAddItem),
@@ -264,14 +265,15 @@ static const char event_type_to_name[][35] =
     EVENT(libvlc_MediaListViewItemDeleted),
     EVENT(libvlc_MediaListViewWillDeleteItem),
 
-    EVENT(libvlc_MediaListPlayerPlayed),
+    EVENT(libvlc_MediaListPlayerStarted),
+    EVENT(libvlc_MediaListPlayerItemPlayed),
     EVENT(libvlc_MediaListPlayerNextItemSet),
+    EVENT(libvlc_MediaListPlayerEndReached),
     EVENT(libvlc_MediaListPlayerStopped),
 
     EVENT(libvlc_MediaDiscovererStarted),
     EVENT(libvlc_MediaDiscovererEnded),
 
-    EVENT(libvlc_MediaPlayerSnapshotTaken),
 #undef EVENT
 };
 
diff --git a/src/control/libvlc_internal.h b/src/control/libvlc_internal.h
index 90fb7a8..06c874f 100644
--- a/src/control/libvlc_internal.h
+++ b/src/control/libvlc_internal.h
@@ -164,9 +164,12 @@ struct libvlc_media_list_player_t
     int                         i_refcount;
     vlc_mutex_t                 object_lock;
     libvlc_media_list_path_t    current_playing_item_path;
-    libvlc_media_t * p_current_playing_item;
     libvlc_media_list_t *       p_mlist;
-    libvlc_media_player_t *   p_mi;
+    libvlc_media_player_t *     p_mi;
+
+    vlc_thread_t                thread;
+    vlc_mutex_t                 thread_lock;
+    vlc_cond_t                  thread_wait;
 };
 
 struct libvlc_media_library_t
diff --git a/src/control/media_list.c b/src/control/media_list.c
index 2eae497..2908f86 100644
--- a/src/control/media_list.c
+++ b/src/control/media_list.c
@@ -495,7 +495,7 @@ void libvlc_media_list_unlock( libvlc_media_list_t * p_mlist )
 
 
 /**************************************************************************
- *       libvlc_media_list_p_event_manager (Public)
+ *       libvlc_media_list_event_manager (Public)
  *
  * The p_event_manager is immutable, so you don't have to hold the lock
  **************************************************************************/
diff --git a/src/control/media_list_path.h b/src/control/media_list_path.h
index c3e698d..5bd5060 100644
--- a/src/control/media_list_path.h
+++ b/src/control/media_list_path.h
@@ -75,7 +75,7 @@ static inline libvlc_media_list_path_t libvlc_media_list_path_copy_by_appending(
     libvlc_media_list_path_t ret;
     int old_deepness = libvlc_media_list_path_deepness( path );
     ret = malloc( sizeof(int)*(old_deepness+2) );
-    memcpy( ret, path, sizeof(int)*(old_deepness+2) );
+    memcpy( ret, path, sizeof(int)*(old_deepness) );
     ret[old_deepness] = index;
     ret[old_deepness+1] = -1;
     return ret;
@@ -106,7 +106,10 @@ get_path_rec( libvlc_media_list_path_t path, libvlc_media_list_t * p_current_mli
         libvlc_media_t * p_md = libvlc_media_list_item_at_index( p_current_mlist, i, NULL );
 
         if( p_md == p_searched_md )
+        {
+            libvlc_media_release( p_md );
             return libvlc_media_list_path_copy_by_appending( path, i ); /* Found! */
+        }
 
         libvlc_media_list_t * p_subitems = libvlc_media_subitems( p_md, NULL );
         libvlc_media_release( p_md );
@@ -149,7 +152,6 @@ libvlc_media_list_item_at_path( libvlc_media_list_t * p_mlist, libvlc_media_list
     for( i = 0; path[i] != -1; i++ )
     {
         p_md = libvlc_media_list_item_at_index( p_current_mlist, path[i], NULL );
-
         if( p_current_mlist != p_mlist )
             libvlc_media_list_release( p_current_mlist );
 
@@ -185,10 +187,12 @@ libvlc_media_list_parentlist_at_path( libvlc_media_list_t * p_mlist, libvlc_medi
             libvlc_media_list_release( p_current_mlist );
 
         if( path[i+1] == -1 )
+        {
+            libvlc_media_list_retain( p_current_mlist );
             return p_current_mlist;
+        }
 
         p_md = libvlc_media_list_item_at_index( p_current_mlist, path[i], NULL );
-
         p_current_mlist = libvlc_media_subitems( p_md, NULL );
         libvlc_media_release( p_md );
  
diff --git a/src/control/media_list_player.c b/src/control/media_list_player.c
index 0f704bd..fe533da 100644
--- a/src/control/media_list_player.c
+++ b/src/control/media_list_player.c
@@ -29,63 +29,147 @@
  */
 
 /**************************************************************************
+ *       media_list_player_thread (private) 
+ **************************************************************************/
+static void *
+media_list_player_thread( void * p_data )
+{
+    libvlc_media_list_player_t * p_mlp = p_data;
+
+    while( 1 )
+    {
+        vlc_mutex_lock( &p_mlp->thread_lock );
+        mutex_cleanup_push( &p_mlp->thread_lock );
+
+        vlc_cond_wait( &p_mlp->thread_wait, &p_mlp->thread_lock );
+
+        vlc_cleanup_run();
+        
+        /* just received signal EndReached from a previous item,
+         * start the next one (if any) */
+        libvlc_media_list_player_next( p_mlp, NULL );
+    }
+}
+
+/**************************************************************************
  *       get_next_index (private)
  *
- * Simple next item fetcher.
+ * next item fetcher.
+ *  - if current item or first item is a node, go to its deepest leaf, play it
+ *  - if not a node and a first item, play it
+ *  - if not a node and not first item, go to next item 
+ *  - if next item is a node, go to its deepest leaf, play it
+ *  - if next item not a node, play it
  **************************************************************************/
 static libvlc_media_list_path_t
 get_next_path( libvlc_media_list_player_t * p_mlp )
 {
-    /* We are entered with libvlc_media_list_lock( p_mlp->p_list ) */
+    /* We are entered with libvlc_media_list_lock( p_mlp->p_mlist ) */
     libvlc_media_list_path_t ret;
     libvlc_media_list_t * p_parent_of_playing_item;
     libvlc_media_list_t * p_sublist_of_playing_item;
 
-    if ( !p_mlp->current_playing_item_path )
+    bool b_first_playing_item;
+    if( !p_mlp->current_playing_item_path )
     {
-        if( !libvlc_media_list_count( p_mlp->p_mlist, NULL ) )
+        /* no current item, try to set up a first item */
+        if( libvlc_media_list_count( p_mlp->p_mlist, NULL ) > 0 )
+        {
+            b_first_playing_item = true;
+            ret = libvlc_media_list_path_with_root_index(0);
+        }
+        else
             return NULL;
-        return libvlc_media_list_path_with_root_index(0);
     }
-    
-    p_sublist_of_playing_item = libvlc_media_list_sublist_at_path(
-                            p_mlp->p_mlist,
-                            p_mlp->current_playing_item_path );
- 
-    /* If item just gained a sublist just play it */
-    if( p_sublist_of_playing_item )
+    else
     {
+        /* current item just played, work from it */
+        b_first_playing_item = false;
+        ret = libvlc_media_list_path_copy( p_mlp->current_playing_item_path );
+    }
+
+    /**
+     * If item (either first item or current item) has got 
+     * sublist(s), just play its deepest leaf 
+     */
+    bool b_sublist_of_playing_item = false;
+    while( 1 )
+    {
+        p_sublist_of_playing_item = libvlc_media_list_sublist_at_path( p_mlp->p_mlist, ret );
+        if( !p_sublist_of_playing_item )
+            break;
+ 
+        libvlc_media_list_path_t ret_tmp;
+        ret_tmp = libvlc_media_list_path_copy_by_appending( ret, 0 );
         libvlc_media_list_release( p_sublist_of_playing_item );
-        return libvlc_media_list_path_copy_by_appending( p_mlp->current_playing_item_path, 0 );
+        free( ret );
+        ret = ret_tmp;
+        b_sublist_of_playing_item = true;
     }
 
-    /* Try to catch next element */
-    p_parent_of_playing_item = libvlc_media_list_parentlist_at_path(
-                            p_mlp->p_mlist,
-                            p_mlp->current_playing_item_path );
+    /* this is a first item or current item with sublists, no need to find more */
+    if( b_first_playing_item || b_sublist_of_playing_item )
+        return ret;
 
+    /**
+     * No sublist for current item, then try to catch next element 
+     * */
     int deepness = libvlc_media_list_path_deepness( p_mlp->current_playing_item_path );
-    if( deepness < 1 || !p_parent_of_playing_item )
-        return NULL;
 
-    ret = libvlc_media_list_path_copy( p_mlp->current_playing_item_path );
-
-    while( ret[deepness-1] >= libvlc_media_list_count( p_parent_of_playing_item, NULL ) )
+    while( deepness > 0 )
     {
-        deepness--;
-        if( deepness <= 0 )
+        if( deepness > 1 )
+            p_parent_of_playing_item = 
+                libvlc_media_list_parentlist_at_path( p_mlp->p_mlist, ret );
+        else
+            p_parent_of_playing_item = p_mlp->p_mlist;
+
+        if( !p_parent_of_playing_item )
         {
             free( ret );
-            libvlc_media_list_release( p_parent_of_playing_item );
             return NULL;
         }
-        ret[deepness] = -1;
+
         ret[deepness-1]++;
-        p_parent_of_playing_item  = libvlc_media_list_parentlist_at_path(
-                                        p_mlp->p_mlist,
-                                        ret );
+        if( ret[deepness-1] < libvlc_media_list_count( p_parent_of_playing_item, NULL ) )
+        {
+            /* found a suitable item */
+            if( deepness > 1 )
+                libvlc_media_list_release( p_parent_of_playing_item );
+            break;
+        }
+
+        if( deepness > 1 )
+            libvlc_media_list_release( p_parent_of_playing_item );
+        deepness--;
+        ret[deepness] = -1;
+    } 
+
+    /* no more item (end of playlist reached) */
+    if( deepness == 0 )
+    {
+        free( ret );
+        return NULL;
+    } 
+
+
+    /**
+     * If new item just found has got a sublist then go for its deepest leaf 
+     */
+    while( 1 )
+    {
+        p_sublist_of_playing_item = libvlc_media_list_sublist_at_path( p_mlp->p_mlist, ret );
+        if( !p_sublist_of_playing_item )
+            break;
+ 
+        libvlc_media_list_path_t ret_tmp;
+        ret_tmp = libvlc_media_list_path_copy_by_appending( ret, 0 );
+        libvlc_media_list_release( p_sublist_of_playing_item );
+        free( ret );
+        ret = ret_tmp;
     }
-    libvlc_media_list_release( p_parent_of_playing_item );
+
+    /* found one */
     return ret;
 }
 
@@ -101,21 +185,39 @@ media_player_reached_end( const libvlc_event_t * p_event,
     libvlc_media_t *p_md, * p_current_md;
 
     p_md = libvlc_media_player_get_media( p_mi, NULL );
-    /* XXX: need if p_mlp->p_current_playing_index is beyond */
+
+    /* check if path is still valid (can return NULL if invalid)  */
     p_current_md = libvlc_media_list_item_at_path(
                         p_mlp->p_mlist,
                         p_mlp->current_playing_item_path );
-    if( p_md != p_current_md )
+    if( !p_current_md || p_current_md != p_md )
     {
         msg_Warn( p_mlp->p_libvlc_instance->p_libvlc_int,
                   "We are not sync-ed with the media instance" );
+
         libvlc_media_release( p_md );
-        libvlc_media_release( p_current_md );
+        if( p_current_md )
+            libvlc_media_release( p_current_md );
+
+        /* used to reset media_list_player and send Stopped event */
+        libvlc_media_list_player_stop( p_mlp, NULL );
+
         return;
     }
-    libvlc_media_release( p_md );
     libvlc_media_release( p_current_md );
-    libvlc_media_list_player_next( p_mlp, NULL );
+
+    /* Send the ItemPlayed event */
+    libvlc_event_t event;
+    event.type = libvlc_MediaListPlayerItemPlayed;
+    event.u.media_item_played.item = p_md;
+    libvlc_event_send( p_mlp->p_event_manager, &event);
+
+    libvlc_media_release( p_md );
+
+    /* Signal the media_list_player thread to start playing the next media */
+    vlc_mutex_lock( &p_mlp->thread_lock );
+    vlc_cond_signal( &p_mlp->thread_wait );
+    vlc_mutex_unlock( &p_mlp->thread_lock );
 }
 
 /**************************************************************************
@@ -124,10 +226,10 @@ media_player_reached_end( const libvlc_event_t * p_event,
 static void
 mlist_item_deleted( const libvlc_event_t * p_event, void * p_user_data )
 {
-    libvlc_media_t * p_current_md;
+    libvlc_media_t * p_current_md = NULL;
     libvlc_media_list_player_t * p_mlp = p_user_data;
     libvlc_media_list_t * p_emitting_mlist = p_event->p_obj;
-    /* XXX: need if p_mlp->p_current_playing_index is beyond */
+ 
     p_current_md = libvlc_media_list_item_at_path(
                         p_mlp->p_mlist,
                         p_mlp->current_playing_item_path );
@@ -141,22 +243,22 @@ mlist_item_deleted( const libvlc_event_t * p_event, void * p_user_data )
 }
 
 /**************************************************************************
- *       install_playlist_observer (private)
+ *       install_media_list_observer (private)
  **************************************************************************/
 static void
-install_playlist_observer( libvlc_media_list_player_t * p_mlp )
+install_media_list_observer( libvlc_media_list_player_t * p_mlp )
 {
     libvlc_event_attach( libvlc_media_list_event_manager( p_mlp->p_mlist, NULL ),
             libvlc_MediaListItemDeleted, mlist_item_deleted, p_mlp, NULL );
 }
 
 /**************************************************************************
- *       uninstall_playlist_observer (private)
+ *       uninstall_media_list_observer (private)
  **************************************************************************/
 static void
-uninstall_playlist_observer( libvlc_media_list_player_t * p_mlp )
+uninstall_media_list_observer( libvlc_media_list_player_t * p_mlp )
 {
-    if ( !p_mlp->p_mlist )
+    if( !p_mlp->p_mlist )
     {
         return;
     }
@@ -183,7 +285,7 @@ install_media_player_observer( libvlc_media_list_player_t * p_mlp )
 static void
 uninstall_media_player_observer( libvlc_media_list_player_t * p_mlp )
 {
-    if ( !p_mlp->p_mi )
+    if( !p_mlp->p_mi )
     {
         return;
     }
@@ -196,7 +298,7 @@ uninstall_media_player_observer( libvlc_media_list_player_t * p_mlp )
 /**************************************************************************
  *       set_current_playing_item (private)
  *
- * Playlist lock should be held
+ * media_list lock should be held
  **************************************************************************/
 static void
 set_current_playing_item( libvlc_media_list_player_t * p_mlp,
@@ -208,43 +310,33 @@ set_current_playing_item( libvlc_media_list_player_t * p_mlp,
     libvlc_media_t * p_md;
 
     p_md = libvlc_media_list_item_at_path( p_mlp->p_mlist, path );
+    if( !p_md )
+        return;
+
     vlc_mutex_lock( &p_mlp->object_lock );
 
-    if( p_mlp->current_playing_item_path != path )
-    {
+    if( p_mlp->current_playing_item_path )
         free( p_mlp->current_playing_item_path );
-        p_mlp->current_playing_item_path = path;
-    }
 
-    if( !p_md )
-    {
-        vlc_mutex_unlock( &p_mlp->object_lock );
-        return;
-    }
+    p_mlp->current_playing_item_path = path;
 
     /* We are not interested in getting media stop event now */
     uninstall_media_player_observer( p_mlp );
 
-    if ( !p_mlp->p_mi )
+    if( !p_mlp->p_mi )
     {
-        p_mlp->p_mi = libvlc_media_player_new_from_media(p_md, p_e);
+        p_mlp->p_mi = libvlc_media_player_new_from_media( p_md, p_e );
     }
-    
-    if( p_md->p_subitems && libvlc_media_list_count( p_md->p_subitems, NULL ) > 0 )
+    else
     {
-        libvlc_media_t * p_submd;
-        p_submd = libvlc_media_list_item_at_index( p_md->p_subitems, 0, NULL );
-        libvlc_media_player_set_media( p_mlp->p_mi, p_submd, NULL );
-        libvlc_media_release( p_submd );
+        libvlc_media_player_set_media( p_mlp->p_mi, p_md, p_e );
     }
-    else
-        libvlc_media_player_set_media( p_mlp->p_mi, p_md, NULL );
-//    wait_playing_state(); /* If we want to be synchronous */
+
     install_media_player_observer( p_mlp );
 
     vlc_mutex_unlock( &p_mlp->object_lock );
 
-    libvlc_media_release( p_md ); /* for libvlc_media_list_item_at_index */
+    libvlc_media_release( p_md ); /* for libvlc_media_list_item_at_path */
 }
 
 /*
@@ -267,12 +359,49 @@ libvlc_media_list_player_new( libvlc_instance_t * p_instance,
     p_mlp->current_playing_item_path = NULL;
     p_mlp->p_mi = NULL;
     p_mlp->p_mlist = NULL;
-    vlc_mutex_init( &p_mlp->object_lock );
+    p_mlp->p_libvlc_instance = p_instance;
+
+    /* event manager init */
     p_mlp->p_event_manager = libvlc_event_manager_new( p_mlp,
                                                        p_instance,
                                                        p_e );
     libvlc_event_manager_register_event_type( p_mlp->p_event_manager,
+            libvlc_MediaListPlayerStarted, p_e );
+    libvlc_event_manager_register_event_type( p_mlp->p_event_manager,
+            libvlc_MediaListPlayerItemPlayed, p_e );
+    libvlc_event_manager_register_event_type( p_mlp->p_event_manager,
             libvlc_MediaListPlayerNextItemSet, p_e );
+    libvlc_event_manager_register_event_type( p_mlp->p_event_manager,
+            libvlc_MediaListPlayerEndReached, p_e );
+    libvlc_event_manager_register_event_type( p_mlp->p_event_manager,
+            libvlc_MediaListPlayerStopped, p_e );
+
+    if( libvlc_exception_raised( p_e ) )
+    {
+        libvlc_event_manager_release( p_mlp->p_event_manager );
+        free( p_mlp );
+        return NULL;
+    }
+
+    vlc_mutex_init( &p_mlp->object_lock );
+
+    /* thread init */
+    vlc_mutex_init( &p_mlp->thread_lock );
+    vlc_cond_init( &p_mlp->thread_wait );
+    if( vlc_clone( &p_mlp->thread, media_list_player_thread, p_mlp,
+                   VLC_THREAD_PRIORITY_LOW ) )
+    {
+        libvlc_exception_raise( p_e, 
+                       "media_list_player thread initialization failed" );
+        vlc_cond_destroy( &p_mlp->thread_wait );
+        vlc_mutex_destroy( &p_mlp->thread_lock );
+  
+        vlc_mutex_destroy( &p_mlp->object_lock );
+
+        libvlc_event_manager_release( p_mlp->p_event_manager );
+        free( p_mlp );
+        return NULL;
+    }
 
     return p_mlp;
 }
@@ -282,7 +411,59 @@ libvlc_media_list_player_new( libvlc_instance_t * p_instance,
  **************************************************************************/
 void libvlc_media_list_player_release( libvlc_media_list_player_t * p_mlp )
 {
-    free(p_mlp);
+    vlc_mutex_lock( &p_mlp->object_lock );
+
+    if( p_mlp->p_mi )
+    {
+        uninstall_media_player_observer( p_mlp );
+        libvlc_media_player_release( p_mlp->p_mi );
+        p_mlp->p_mi = NULL;
+    }
+
+    if( p_mlp->p_mlist )
+    {
+        uninstall_media_list_observer( p_mlp );
+        libvlc_media_list_release( p_mlp->p_mlist );
+        p_mlp->p_mlist = NULL;
+    }
+
+    vlc_mutex_unlock( &p_mlp->object_lock );
+
+    libvlc_event_manager_release( p_mlp->p_event_manager );
+
+    vlc_cancel( p_mlp->thread );
+    vlc_join( p_mlp->thread, NULL );
+
+    vlc_cond_destroy( &p_mlp->thread_wait );
+    vlc_mutex_destroy( &p_mlp->thread_lock );
+
+    vlc_mutex_destroy( &p_mlp->object_lock );
+
+    free( p_mlp );
+}
+
+/**************************************************************************
+ *        get_media_player (Public)
+ **************************************************************************/
+libvlc_media_player_t *
+libvlc_media_list_player_get_media_player(
+                                     libvlc_media_list_player_t * p_mlp,
+                                     libvlc_exception_t * p_e )
+{
+    VLC_UNUSED(p_e);
+    libvlc_media_player_t * p_mi = NULL;
+
+    vlc_mutex_lock( &p_mlp->object_lock );
+
+    if( p_mlp->p_mi )
+    {
+        p_mi = p_mlp->p_mi;
+        libvlc_media_player_retain( p_mi );
+    }
+
+    vlc_mutex_unlock( &p_mlp->object_lock );
+  
+    return p_mi;
 }
 
 /**************************************************************************
@@ -302,10 +483,14 @@ void libvlc_media_list_player_set_media_player(
         uninstall_media_player_observer( p_mlp );
         libvlc_media_player_release( p_mlp->p_mi );
     }
-    libvlc_media_player_retain( p_mi );
+
     p_mlp->p_mi = p_mi;
 
-    install_media_player_observer( p_mlp );
+    if ( p_mi )
+    {
+        libvlc_media_player_retain( p_mi );
+        install_media_player_observer( p_mlp );
+    }
 
     vlc_mutex_unlock( &p_mlp->object_lock );
 }
@@ -322,21 +507,25 @@ void libvlc_media_list_player_set_media_list(
 
     if( libvlc_media_list_player_is_playing( p_mlp, p_e ) )
     {
-        libvlc_media_player_stop( p_mlp->p_mi, p_e );
+        libvlc_media_list_player_stop( p_mlp, p_e );
         /* Don't bother if there was an error. */
         libvlc_exception_clear( p_e );
     }
 
     if( p_mlp->p_mlist )
     {
-        uninstall_playlist_observer( p_mlp );
+        uninstall_media_list_observer( p_mlp );
         libvlc_media_list_release( p_mlp->p_mlist );
     }
-    libvlc_media_list_retain( p_mlist );
+
     p_mlp->p_mlist = p_mlist;
- 
-    install_playlist_observer( p_mlp );
 
+    if ( p_mlist )
+    {
+        libvlc_media_list_retain( p_mlist );
+        install_media_list_observer( p_mlp );
+    }
+ 
     vlc_mutex_unlock( &p_mlp->object_lock );
 }
 
@@ -348,8 +537,9 @@ void libvlc_media_list_player_play( libvlc_media_list_player_t * p_mlp,
 {
     if( !p_mlp->current_playing_item_path )
     {
+        /* Will set to play */
         libvlc_media_list_player_next( p_mlp, p_e );
-        return; /* Will set to play */
+        return;
     }
 
     libvlc_media_player_play( p_mlp->p_mi, p_e );
@@ -400,16 +590,22 @@ void libvlc_media_list_player_play_item_at_index(
                         libvlc_exception_t * p_e )
 {
     set_current_playing_item( p_mlp, libvlc_media_list_path_with_root_index(i_index), p_e );
+    if( libvlc_exception_raised( p_e ) )
+        return;
 
+    libvlc_media_t* p_md = libvlc_media_list_item_at_index( p_mlp->p_mlist, i_index, p_e );
     if( libvlc_exception_raised( p_e ) )
         return;
 
+    libvlc_media_player_play( p_mlp->p_mi, p_e );
+
     /* Send the next item event */
     libvlc_event_t event;
     event.type = libvlc_MediaListPlayerNextItemSet;
+    event.u.media_next_item_set.item = p_md;
     libvlc_event_send( p_mlp->p_event_manager, &event );
+    libvlc_media_release( p_md );
 
-    libvlc_media_player_play( p_mlp->p_mi, p_e );
 }
 
 /**************************************************************************
@@ -440,17 +636,25 @@ void libvlc_media_list_player_play_item(
 void libvlc_media_list_player_stop( libvlc_media_list_player_t * p_mlp,
                                     libvlc_exception_t * p_e )
 {
-    if ( p_mlp->p_mi )
+    if( p_mlp->p_mi )
     {
         /* We are not interested in getting media stop event now */
         uninstall_media_player_observer( p_mlp );
         libvlc_media_player_stop( p_mlp->p_mi, p_e );
         install_media_player_observer( p_mlp );
+
+        /* send Stopped event */
+        libvlc_event_t event;
+        event.type = libvlc_MediaListPlayerStopped;
+        libvlc_event_send( p_mlp->p_event_manager, &event);
     }
 
     vlc_mutex_lock( &p_mlp->object_lock );
-    free( p_mlp->current_playing_item_path );
-    p_mlp->current_playing_item_path = NULL;
+    if( p_mlp->current_playing_item_path )
+    {
+        free( p_mlp->current_playing_item_path );
+        p_mlp->current_playing_item_path = NULL;
+    }
     vlc_mutex_unlock( &p_mlp->object_lock );
 }
 
@@ -462,12 +666,20 @@ void libvlc_media_list_player_next( libvlc_media_list_player_t * p_mlp,
 {
     libvlc_media_list_path_t path;
 
-    if (! p_mlp->p_mlist )
+    if( !p_mlp->p_mlist )
     {
-        libvlc_exception_raise( p_e, "No more element to play" );
+        libvlc_exception_raise( p_e, "No media list available" );
         return;
     }
 
+    if( !p_mlp->current_playing_item_path )
+    {
+        /* Send the Started event */
+        libvlc_event_t event;
+        event.type = libvlc_MediaListPlayerStarted;
+        libvlc_event_send( p_mlp->p_event_manager, &event);
+    }
+
     libvlc_media_list_lock( p_mlp->p_mlist );
 
     path = get_next_path( p_mlp );
@@ -475,7 +687,12 @@ void libvlc_media_list_player_next( libvlc_media_list_player_t * p_mlp,
     if( !path )
     {
         libvlc_media_list_unlock( p_mlp->p_mlist );
-        libvlc_exception_raise( p_e, "No more element to play" );
+
+        /* Send the EndReached item event */
+        libvlc_event_t event;
+        event.type = libvlc_MediaListPlayerEndReached;
+        libvlc_event_send( p_mlp->p_event_manager, &event);
+
         libvlc_media_list_player_stop( p_mlp, p_e );
         return;
     }
@@ -487,7 +704,23 @@ void libvlc_media_list_player_next( libvlc_media_list_player_t * p_mlp,
     libvlc_media_list_unlock( p_mlp->p_mlist );
 
     /* Send the next item event */
+    libvlc_media_t* p_md = libvlc_media_player_get_media( p_mlp->p_mi, NULL );
     libvlc_event_t event;
     event.type = libvlc_MediaListPlayerNextItemSet;
+    event.u.media_next_item_set.item = p_md;
     libvlc_event_send( p_mlp->p_event_manager, &event);
+    libvlc_media_release( p_md );
+}
+
+/**************************************************************************
+ *       libvlc_media_list_player_event_manager (Public)
+ *
+ * The p_event_manager is immutable, so you don't have to hold the lock
+ **************************************************************************/
+libvlc_event_manager_t *
+libvlc_media_list_player_event_manager( libvlc_media_list_player_t * p_mlp,
+                                        libvlc_exception_t * p_e )
+{
+    (void)p_e;
+    return p_mlp->p_event_manager;
 }
diff --git a/src/control/media_player.c b/src/control/media_player.c
index 8cd0cef..777c405 100644
--- a/src/control/media_player.c
+++ b/src/control/media_player.c
@@ -114,7 +114,7 @@ input_thread_t *libvlc_get_input_thread( libvlc_media_player_t *p_mi,
     if( !p_mi->p_input_thread )
     {
         vlc_mutex_unlock( &p_mi->object_lock );
-        RAISENULL( "Input is NULL" );
+        return NULL;
     }
 
     p_input_thread = p_mi->p_input_thread;
diff --git a/src/libvlc.sym b/src/libvlc.sym
index 4c7efb1..29252fb 100644
--- a/src/libvlc.sym
+++ b/src/libvlc.sym
@@ -91,8 +91,10 @@ libvlc_media_list_player_play_item
 libvlc_media_list_player_play_item_at_index
 libvlc_media_list_player_release
 libvlc_media_list_player_set_media_list
+libvlc_media_list_player_get_media_player
 libvlc_media_list_player_set_media_player
 libvlc_media_list_player_stop
+libvlc_media_list_player_event_manager
 libvlc_media_list_release
 libvlc_media_list_remove_index
 libvlc_media_list_retain
-- 
1.5.2.5

